#!/usr/bin/env python
# -*- coding: iso-8859-1 -*-
"""
    GESTIKK
    Version 0.2b

    Gestikk, a mouse gesture recognition application
    Copyright (C) 2007 Friedrich 'fred.reichbier' Weber (2007-10-08)

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

from __future__ import division
import os.path
import wx
import time
import gettext
import locale
import subprocess
import ConfigParser
from thread import start_new_thread

if not os.path.isdir(os.path.expanduser('~/.gestikk')): # Create directory is it doesnt exist
    os.mkdir(os.path.expanduser('~/.gestikk'))
CONFIG_FILE = os.path.expanduser('~/.gestikk/gestures.conf') # TODO

wx.SetDefaultPyEncoding("iso-8859-1")

def calc_direction(pos1, pos2, shw=100):
    """ Calculate direction from pos1 to pos2, return tuple (x,y), e.g. (1,1
        Shw is the 'threshold value', as higher as more tolerant. recommended: ~ 100 """
    move_x, move_y = 0,0
    diff = (pos2[0] - pos1[0], pos2[1] - pos1[1])
    if abs(diff[0]) >= shw:
        # moved --
        if diff[0] > 0:
            move_x = 1
        elif diff[0] < 0:
            move_x = -1
        else:
            move_x = 0
    if abs(diff[1]) >= shw:
        # moved |
        if diff[1] > 0:
            move_y = 1
        elif diff[1] < 0:
            move_y = -1
        else:
            move_y = 0
    return (move_x, move_y)

def get_language():
    """ get the preferred language, e.g. de_DE """
    cp = ConfigParser.ConfigParser() # nicht gerade sauber, aber sonst fllt mir nichts ein
    cp.read(CONFIG_FILE)
    try:
        if cp.has_section('Preferences') \
            and cp.has_option('Preferences', 'language') \
            and cp.get('Preferences', 'language') != '':
            return cp.get('Preferences', 'language')
        else:
            return locale.getdefaultlocale()[0]
    finally:
        del cp

def init_gettext():
    """ init gettext """
    gettext.install('config')
    lang = gettext.translation('config', 'locale', languages=[get_language()], fallback=True)
    return lang

_ = init_gettext().gettext

def my_round(i, shw=25):
    """ Not really round but use divmod. Shw is the 'threshold value', as higher as more tolerant. threshold value = 5*Value """
    pow10 = 5*shw
    return (i // pow10) * pow10

def remove_duplicates(l):
    """ Remove upfollowing duplicates from the list """
    new= l[:]
    last_e = None
    for e in l:
        if e == last_e:
            new.remove(e)
        last_e = e
    return new

class ActionDialog(wx.Dialog):
    """ Choose Action - Dialogue """
    def __init__(self, logic, parent=None, id=-1, title=_('Action'), size=wx.Size(550, 430)):
        wx.Dialog.__init__(self, parent, id, title, size=size)
        self.logic = logic

        sizer = wx.BoxSizer(wx.VERTICAL)
        lbl_help = wx.StaticText(self, label=_('Choose now which action should be executed.'))
        sizer.Add(lbl_help)
        panel = wx.Panel(self)
        # now the interestin part #
        grid_sizer = wx.GridSizer(rows=2, cols=2)
        # list
        self.list_keys = self.logic.action_list.keys()
        lst_actions = wx.ListBox(panel, choices=self.list_keys, size=wx.Size(150,200))
        grid_sizer.Add(lst_actions)
        self.lst_actions = lst_actions
        lst_actions.Bind(wx.EVT_LISTBOX, self.evt_list)
        # # # # # # #
        sizer2 = wx.BoxSizer(wx.VERTICAL)
        panel2 = wx.Panel(panel)
        # input box
        input = wx.TextCtrl(panel2, size=wx.Size(200, 25))
        self.input = input
        sizer2.Add(input, 0, wx.ALIGN_RIGHT | wx.ALL)
        # helptext
        lbl_help = wx.StaticText(panel2, size=wx.Size(200,300))
        self.lbl_help = lbl_help
        sizer2.Add(lbl_help, 0, wx.ALL, 5)
        # # # # # # #
        panel2.SetSizer(sizer2)
        # # # # # # #
        grid_sizer.Add(panel2)
        # # # # # # # # # # # # # #
        panel.SetSizer(grid_sizer)
        # # # #
        button = wx.Button(self, label=_('Create'))
        button.Bind(wx.EVT_BUTTON, self.evt_save)
        sizer.Add(panel, 0, wx.ALL, 5)
        sizer.Add(button)
        # #
        self.SetSizer(sizer)

        self.current_template = None

    def evt_save(self, evt):
        """ save """
        if self.current_template:
            self.value = self.current_template.get(self.input.GetValue())
        self.EndModal(wx.ID_OK)

    def evt_list(self, evt):
        """ something new chosen, show helptext """
        try:
            template = self.logic.action_list[self.list_keys[self.lst_actions.GetSelections()[0]]]
            self.current_template = template
        except IndexError:
            print _('Small error, that hasn\'t much to say (1)')
            return
        # show whelptext
        self.lbl_help.SetLabel(template.help)
        self.lbl_help.Wrap(200)
        evt.Skip()


class NewGestureDialog(wx.Dialog):
    """ Neue Mausgeste - Dialog """
    def __init__(self, logic, parent=None, id=-1, title=_('New Gesture'), size=wx.Size(550, 430)):
        wx.Dialog.__init__(self, parent, id, title, size=size)
        self.logic = logic

        sizer = wx.BoxSizer(wx.VERTICAL)
        lbl_do = wx.StaticText(self, label=_('Draw the gesture (big one!) in the window while pressing the right mouse button.\nAttention! An already drawed action will be overwritten!'))
        sizer.Add(lbl_do)
        self.lbl_do = lbl_do

        self.gesture_positions = []

        self.Bind(wx.EVT_CLOSE, self.evt_close)
        self.is_looping = 1 # is set to 0 if thread should be closed
        self.gestures = []
        start_new_thread(self.mouse_gesture_thread, ())

    def evt_close(self, evt):
        """ Closing: Clsoe thread if not already closed (like an emergency brake :-)) """
        self.is_looping = 0
        time.sleep(0.26)
        evt.Skip()

    def mouse_gesture_thread(self):
        """ Thread for mouse gesture recognition """
        while self.is_looping:
            wx.CallAfter(wx.YieldIfNeeded)
            wx.CallAfter(self.check_mouse_position)
            time.sleep(0.25)

    def check_gestures(self):
        """ Check both gestures in self.gestures """
        if self.gestures[0] != self.gestures[1]: # Bot the same
            self.lbl_do.SetLabel(_('Unfortunately the gestures don\'t match.\nTry again please!'))
            self.gestures = []
        else:
            #dlg = wx.TextEntryDialog(
            #        self, _('Now enter the command that will be executed by this gesture (eg. gksu gimp).'),
            #        _('Aktion'), '')
            dlg = ActionDialog(self.logic, self)
            if dlg.ShowModal() == wx.ID_OK:
                self.logic.actions[self.gestures[0][1:-1]] = dlg.value # We need the gestures without edgy brackets to not confuse the ConfigParser
                wx.CallAfter(self.Close)
            else:
                self.lbl_do.SetLabel(_('If you want to, try again.\nElse just close this dialogue.'))
                self.gestures = []
            dlg.Destroy()


    def check_mouse_position(self):
        """ Check position of mouse pointer """
        pos = wx.GetMousePosition()
        mstate = wx.GetMouseState().RightDown()
        if mstate: # gesture because right button is pressed
            self.gesture_positions.append(pos)
        else: # Not longer pressed
            if self.gesture_positions: # gesture finished
                # Pfad errechnen
                gesture = []
                tolerance = int(self.logic.config_get_option('Preferences', 'tolerance', '25'))
                for gid in range(1, len(self.gesture_positions)): # let first out
                    gesture.append(calc_direction(self.gesture_positions[gid-1], \
                                                self.gesture_positions[gid], tolerance))
                # Remove zeros
                while (0,0) in gesture:
                    gesture.remove((0,0))
                # Delete duplicates
                gesture = remove_duplicates(gesture)
                self.gesture_positions = []
                # Is a command assigned to the gesture?
                gesture_repr = repr(gesture).replace(' ', '') # w/o spaces
                self.gestures.append(gesture_repr)
                if gesture_repr == '[]': # leer?
                    self.lbl_do.SetLabel(_('That was too fast.\nPlease do it again.'))
                    self.gestures = []
                    self.gesture_positions = []
                    return
                if len(self.gestures) == 1:
                    # first gesture: encourage the user
                    self.lbl_do.SetLabel(_('Now again please so that I can see,\nif you misdrawed it :-)'))
                else:
                    # Second gesture finished, check it
                    self.check_gestures()
                print 'Gesture: %s' % gesture_repr

class ActionTemplate():
    """ Template class for writin actions into configuration file """
    def __init__(self, ident, help):
        """ Ident is the Identifier before the ! (cmd!gksu scite)
            help is a helptext"""
        self.ident = ident
        self.help = help

    def get(self, parameter):
        """ Fetch line from INI-File with parameters """
        return self.ident+'!'+parameter

class Logic():
    """ class for saving the logic (of gestures and preferences) """
    def __init__(self):
        self.actions = {}
        self.config = ConfigParser.ConfigParser()
        self.config_file = CONFIG_FILE # todo
        self.config.read(self.config_file)

        self.actions = self.config_get_gestures()

        self.action_list = {_('Command'):
            ActionTemplate('cmd', _('Execute a command (eg. gksu scite)')),
            _('Keypress'):
            ActionTemplate('key', _('Press a key combination (eg. alt+Left)'))}

    def config_get_option(self, section, option, default=''):
        """ Fetch option """
        if self.config.has_section(section) and self.config.has_option(section, option):
            return self.config.get(section, option)
        else:
            return default

    def config_get_gestures(self):
        """ Fetch all gestures; Dictionary {Gesture:Action, ...} """
        if not self.config.has_section('Gestures'):
            return {}
        dic = {}
        for item in self.config.items('Gestures'): # [(1,2), (3,4)] zu {1:2, 3:4}
            dic[item[0]] = item[1]
        return dic


    def config_set_option(self, section, option, value):
        """ Setze Option, lege ggf. Sektion neu an """
        if not self.config.has_section(section):
            self.config.add_section(section)
        self.config.set(section, option, value)

    def config_write(self):
        """ Write configuration file """
        fp = file(self.config_file, 'w')
        self.config.write(fp)
        fp.close()


class MGConfigFrame(wx.Frame):
    def __init__(self, logic, parent = None, id = -1, title = _("Gestikk - Configuration"), size=wx.Size(640,480)):
        wx.Frame.__init__(self, parent, id, title, size=size)
        self.logic = logic
        # Setze Values-Liste
        #self.values = {} # this is what later will be written to the config file
        #                 # Contains: {Control: (Option, Modificator
        sizer = wx.BoxSizer(wx.VERTICAL)
        #sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(sizer)
        self.notebook = wx.Notebook(self)
        sizer.Add(self.notebook, proportion=1, flag=wx.ALL|wx.EXPAND) # dont know if this is correct but it works
        # 'Save' - Button
        btnsave = wx.Button(self, label=_('Save'))
        btnsave.Bind(wx.EVT_BUTTON, self.save_all)
        sizer.Add(btnsave)
        # Seite 1:
        self.notebook.AddPage(self.get_page_preferences(), _('Preferences'))
        self.notebook.AddPage(self.get_page_gestures(), _('Gestures'))


    def get_option(self, section, option, default=''):
        """ Wrapper for self.logic.config_get_option """
        return self.logic.config_get_option(section, option, default)

    def save_preferences(self):
        """ Save the preferences """
        self.logic.config_set_option('Preferences', 'tolerance', self.page_preferences.v_tolerance.GetValue())
        self.logic.config_set_option('Preferences', 'language', self.page_preferences.v_lang.GetValue())

    def save_gestures(self):
        """ Save the gestures """
        # firstly remove
        self.logic.config.remove_section('Gestures')
        self.logic.config.add_section('Gestures')
        for gesture, action in self.logic.actions.iteritems():
            self.logic.config.set('Gestures', gesture, action)

    def save_all(self, evt):
        """ Save everything """
        self.save_preferences()
        self.save_gestures()
        self.logic.config_write()

    def get_page_preferences(self):
        """ Put out panel 'Preferences' """
        #--- Sizer u.a.
        page = wx.Panel(self.notebook)
        sizer = wx.GridSizer(rows=1, cols=2)
        page.SetSizer(sizer)
        #--- Inhalt
        # - Toleranz
        lbl_tolerance = wx.StaticText(page, label=_('Tolerance (as higher as more tolerant)\nAttention: A change could mean,\nthat the gesture doesnt work anymore.'))
        sizer.Add(lbl_tolerance)
        tolerance = wx.SpinCtrl(page, initial=int(self.get_option('Preferences', 'tolerance', '25')))
        sizer.Add(tolerance)
        page.v_tolerance = tolerance
        # - Sprache
        lbl_lang = wx.StaticText(page, label=_('Default Language (e.g. de_DE)\nIf empty, read it from system\'s default'))
        sizer.Add(lbl_lang)
        lang = wx.TextCtrl(page)
        lang.SetValue(self.get_option('Preferences', 'language', ''))
        sizer.Add(lang)
        page.v_lang = lang
        #--- Fertig
        self.page_preferences = page
        return page

    def fill_gesture_list(self):
        """ Fill list of gestures """
        self.page_gestures.v_glist.Clear()
        i = 0
        for (gesture, action) in self.logic.actions.iteritems():
            self.page_gestures.v_glist.Append(action)
            self.page_gestures.v_glist.SetClientData(i, gesture)
            i += 1 # ye, strange workaround, but i had no wxPython-Dokumentation by hand

    def get_page_gestures(self):
        """ Put out panel 'Gestures' """
         #--- Sizer and others
        page = wx.Panel(self.notebook)
        sizer = wx.GridSizer(rows=1, cols=2)
        page.SetSizer(sizer)
        #--- Content
        # - list
        lbl_glist = wx.StaticText(page, label=_('List of all gestures'))
        sizer.Add(lbl_glist)
        glist = wx.ListBox(page, size=wx.Size(200,200))
        # fill list
        sizer.Add(glist)
        page.v_glist = glist
        self.page_gestures = page
        self.fill_gesture_list()

        # - new gesture
        new_gesture = wx.Button(page, label=_('New gesture'))
        sizer.Add(new_gesture)
        new_gesture.Bind(wx.EVT_BUTTON, self.evt_new_gesture)
        # - Delete gesture
        del_gesture = wx.Button(page, label=_('Delete gesture'))
        sizer.Add(del_gesture)
        del_gesture.Bind(wx.EVT_BUTTON, self.evt_del_gesture)
        #--- Ready

        return page

    def evt_new_gesture(self, evt):
        """ New gesture """
        diag = NewGestureDialog(self.logic)
        #diag = ActionDialog(self.logic)
        diag.ShowModal()
        # Update the list
        self.fill_gesture_list()
        diag.Destroy()

    def evt_del_gesture(self, evt):
        """ delete gesture """
        gesture = self.page_gestures.v_glist.GetClientData(self.page_gestures.v_glist.GetSelection())
        if gesture:
            # delete
            del self.logic.actions[gesture]
            self.fill_gesture_list()





def main():
    app = wx.PySimpleApp()
    logic = Logic()
    diag = MGConfigFrame(logic)
    diag.Show()
    app.MainLoop()


if __name__ == "__main__":
    main()
